//  ************************************************************************************************
//
//  BornAgain: simulate and fit reflection and scattering
//
//! @file      Sim/Simulation/ISimulation.h
//! @brief     Defines interface ISimulation.
//!
//! @homepage  http://www.bornagainproject.org
//! @license   GNU General Public License v3 or higher (see COPYING)
//! @copyright Forschungszentrum Jülich GmbH 2018
//! @authors   Scientific Computing Group at MLZ (see CITATION, AUTHORS)
//
//  ************************************************************************************************

#ifndef BORNAGAIN_SIM_SIMULATION_ISIMULATION_H
#define BORNAGAIN_SIM_SIMULATION_ISIMULATION_H

#include "Param/Distrib/DistributionHandler.h"
#include "Param/Distrib/ParameterDistribution.h"
#include "Param/Node/INode.h"
#include <heinz/Vectors3D.h>

class IBackground;
class IComputation;
class ICoordSystem;
class MultiLayer;
class ProgressHandler;
class Datafield;
class SimulationOptions;
class SimulationResult;
class ReSample;

//! Abstract base class, holds the infrastructure to run a simulation.
//!
//! Base class of SpecularSimulation, ISimulation2D and DepthProbeSimulation.
//!
//! Holds an instrument and sample model. Provides the common infrastructure
//! to run a simulation: multithreading, batch processing, averaging over
//! parameter distributions, etc.
//!
//! Simulations are run, and results returned, by the function ISimulation::simulate().

class ISimulation : public INode {
public:
    ISimulation(const MultiLayer& sample);
    ~ISimulation() override;

    ISimulation(const ISimulation&) = delete;
    ISimulation(ISimulation&&) = default;

    //! Run a simulation, and return the result.
    SimulationResult simulate();

    void setBackground(const IBackground& bg);

    void addParameterDistribution(ParameterDistribution::WhichParameter whichParameter,
                                  const IDistribution1D& distribution, size_t nbr_samples,
                                  double sigma_factor = 0.0,
                                  const RealLimits& limits = RealLimits());
    void addParameterDistribution(const ParameterDistribution& par_distr);

    const SimulationOptions& options() const;
    SimulationOptions& options();

    void setTerminalProgressMonitor();

    //! Force polarized computation even in absence of sample magnetization or external fields
    virtual bool force_polarized() const = 0;

#ifndef SWIG
    void subscribe(const std::function<bool(size_t)>& inform);

    std::string unitOfParameter(ParameterDistribution::WhichParameter which) const;

    const MultiLayer* sample() const;

    const IBackground* background() const
    {
        return m_background.get();
    }

    const std::vector<ParameterDistribution>& getDistributions() const;

    std::vector<const INode*> nodeChildren() const override;

    virtual const ICoordSystem* createCoordSystem() const = 0;

protected:
    ProgressHandler& progress();

    DistributionHandler m_distribution_handler;

    virtual void initDistributionHandler() {}

private:
    void runSingleSimulation(const ReSample& re_sample, size_t batch_start, size_t batch_size,
                             double weight = 1.0);

    //! Put into a clean state for running a simulation
    virtual void prepareSimulation() = 0;

    virtual SimulationResult pack_result() const = 0;

    //! Initializes the vector of ISimulation elements
    virtual void initElementVector() = 0;

    //! Gets the number of elements this simulation needs to calculate
    virtual size_t numberOfElements() const = 0;

    //! Generate a single threaded computation for a given range of simulation elements
    //! @param re_sample Preprocessed version of our sample
    //! @param start Index of the first element to include into computation
    //! @param n_elements Number of elements to process
    virtual std::unique_ptr<IComputation> createComputation(const ReSample& re_sample, size_t start,
                                                            size_t n_elements) = 0;

    //! Checks the distribution validity for simulation.
    virtual void validateParametrization(const ParameterDistribution&) const {}

    virtual void addBackgroundIntensity(size_t start_ind, size_t n_elements) = 0;

    //! Normalize the detector counts to beam intensity, to solid angle, and to exposure angle.
    //! @param start_ind Index of the first element to operate on
    //! @param n_elements Number of elements to process
    virtual void normalize(size_t start_ind, size_t n_elements) = 0;

    virtual void addDataToCache(double weight) = 0;

    virtual void moveDataFromCache() = 0;

    std::shared_ptr<MultiLayer> m_sample;

    std::shared_ptr<SimulationOptions> m_options;
    std::shared_ptr<ProgressHandler> m_progress;
    std::shared_ptr<IBackground> m_background;
#endif // SWIG
};

#endif // BORNAGAIN_SIM_SIMULATION_ISIMULATION_H
